# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.test_files
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    tests for file based handlers

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import socketio
import requests
import json
import os
from functools import partial
import time
import sys
import pytest

from PyQt5 import QtWidgets

from werkzeug.urls import url_join

from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.models import db, User, Project, Change, Permission, Message
from mslib.mscolab.utils import get_recent_pid
from mslib._tests.utils import mscolab_register_and_login, mscolab_create_project, mscolab_start_server
from mslib.msui.mscolab import MSSMscolabWindow


PORTS = list(range(9361, 9380))


@pytest.mark.skipif(os.name == "nt",
                    reason="multiprocessing needs currently start_method fork")
class Test_Files(object):
    def setup(self):
        self.process, self.url, self.app, _, self.cm, self.fm = mscolab_start_server(PORTS)
        time.sleep(0.1)
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = MSSMscolabWindow(data_dir=mscolab_settings.MSCOLAB_DATA_DIR,
                                       mscolab_server_url=self.url)
        self.sockets = []
        self.file_message_counter = [0] * 2
        with self.app.app_context():
            self.user = User.query.filter_by(id=8).first()

    def teardown(self):
        for socket in self.sockets:
            socket.disconnect()
        if self.window.version_window:
            self.window.version_window.close()
        if self.window.conn:
            self.window.conn.disconnect()
        self.application.quit()
        QtWidgets.QApplication.processEvents()
        self.process.terminate()

    def test_create_project(self):
        with self.app.app_context():
            # test for blank character in path
            assert self.fm.create_project('test path', 'test desc.', self.user) is False
            # test for normal path
            assert self.fm.create_project('test_path', 'test desc.', self.user) is True
            # test for '/' in path
            assert self.fm.create_project('test/path', 'sth', self.user) is False
            # check file existence
            assert os.path.exists(os.path.join(mscolab_settings.MSCOLAB_DATA_DIR, 'test_path')) is True
            # check creation in db
            p = Project.query.filter_by(path="test_path").first()
            assert p is not None
            # check permission for author
            perms = Permission.query.filter_by(p_id=p.id, access_level="creator").all()
            assert len(perms) == 1
            assert perms[0].u_id == self.user.id

    def test_projects(self):
        with self.app.app_context():
            projects = self.fm.list_projects(self.user)
            assert len(projects) == 3

    def test_is_admin(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            u_id = self.user.id
            assert self.fm.is_admin(u_id, p_id) is True
            undefined_p_id = 123
            assert self.fm.is_admin(u_id, undefined_p_id) is False
            no_perm_p_id = 2
            assert self.fm.is_admin(u_id, no_perm_p_id) is False

    def test_file_save(self):
        pytest.skip('timing problem')
        url = url_join(self.url, 'token')
        r = requests.post(url, data={
            'email': 'a',
            'password': 'a'
        })
        response1 = json.loads(r.text)
        r = requests.post(url, data={
            'email': 'b',
            'password': 'b'
        })
        response2 = json.loads(r.text)

        def handle_chat_message(sno, message):
            self.file_message_counter[sno - 1] += 1

        sio1 = socketio.Client()
        self.sockets.append(sio1)
        sio2 = socketio.Client()
        self.sockets.append(sio2)

        sio1.on('file-changed', handler=partial(handle_chat_message, 1))
        sio2.on('file-changed', handler=partial(handle_chat_message, 2))
        sio1.connect(self.url)
        sio2.connect(self.url)
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            user2 = User.query.filter_by(id=9).first()
            perm = Permission(u_id=9, p_id=p_id, access_level="admin")
            db.session.add(perm)
            db.session.commit()
            sio1.emit('start', response1)
            sio2.emit('start', response2)
            time.sleep(0.1)
            sio1.emit('file-save', {
                      "p_id": p_id,
                      "token": response1['token'],
                      "content": "file save content 1"
                      })
            time.sleep(0.1)
            # second file change
            sio1.emit('file-save', {
                      "p_id": p_id,
                      "token": response1['token'],
                      "content": "file save content 2"
                      })
            time.sleep(0.1)
            # check if there were events triggered related to file-save
            assert self.file_message_counter[0] == 2
            assert self.file_message_counter[1] == 2
            # check if content is saved in file
            assert self.fm.get_file(p_id, user2) == "file save content 2"
            # check if change is saved properly
            changes = self.fm.get_all_changes(p_id, self.user)
            assert len(changes) == 2
            change = Change.query.order_by(Change.created_at.desc()).first()
            change_content = self.fm.get_change_content(change.id)
            assert change_content == "file save content 2"

    def test_undo(self):
        url = url_join(self.url, 'token')
        r = requests.post(url, data={
            'email': 'a',
            'password': 'a'
        })
        response1 = json.loads(r.text)
        r = requests.post(url, data={
            'email': 'b',
            'password': 'b'
        })
        response2 = json.loads(r.text)

        def handle_chat_message(sno, message):
            self.file_message_counter[sno - 1] += 1

        sio1 = socketio.Client()
        self.sockets.append(sio1)
        sio2 = socketio.Client()
        self.sockets.append(sio2)

        sio1.on('file-changed', handler=partial(handle_chat_message, 1))
        sio2.on('file-changed', handler=partial(handle_chat_message, 2))
        sio1.connect(self.url)
        sio2.connect(self.url)
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            User.query.filter_by(id=9).first()
            perm = Permission(u_id=9, p_id=p_id, access_level="admin")
            db.session.add(perm)
            db.session.commit()
            sio1.emit('start', response1)
            sio2.emit('start', response2)
            time.sleep(0.1)
            sio1.emit('file-save', {
                "p_id": p_id,
                "token": response1['token'],
                "content": "file save content 1"
            })
            time.sleep(0.1)
            # second file change
            sio1.emit('file-save', {
                "p_id": p_id,
                "token": response1['token'],
                "content": "file save content 2"
            })
            time.sleep(0.1)
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            changes = Change.query.filter_by(p_id=p_id).all()
            assert self.fm.undo(changes[0].id, self.user) is True
            assert len(self.fm.get_all_changes(p_id, self.user)) == 3
            assert self.fm.get_file(p_id, self.user) == "file save content 1"

    def test_get_project(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            assert self.fm.get_file(p_id, self.user) is not False
            user2 = User.query.filter_by(id=9).first()
            assert self.fm.get_file(p_id, user2) is False

    def test_authorized_users(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            assert p_id == 4
            assert len(self.fm.get_authorized_users(p_id)) == 2

    def test_modify_project(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            # testing for wrong characters in path like ' ', '/'
            assert self.fm.update_project(p_id, 'path', 'dummy wrong', self.user) is False
            assert self.fm.update_project(p_id, 'path', 'dummy/wrong', self.user) is False
            assert self.fm.update_project(p_id, 'path', 'dummy', self.user) is True
            assert os.path.exists(os.path.join(mscolab_settings.MSCOLAB_DATA_DIR, 'dummy'))
            assert self.fm.update_project(p_id, 'description', 'dummy', self.user) is True

    def test_delete_project(self):
        with self.app.app_context():
            response = mscolab_register_and_login(self.app, self.url, 'a', 'a', 'a')
            data, response = mscolab_create_project(self.app, self.url, response,
                                                    path='f3', description='f3 test example')
            p_id = get_recent_pid(self.fm, self.user)
            assert p_id == 7
            assert response.status == '200 OK'
            user2 = User.query.filter_by(id=9).first()
            assert self.fm.delete_file(p_id, user2) is False
            assert self.fm.delete_file(p_id, self.user) is True
            assert self.fm.delete_file(p_id, self.user) is False
            permissions = Permission.query.filter_by(p_id=p_id).all()
            assert len(permissions) == 0
            projects_db = Project.query.filter_by(id=p_id).all()
            assert len(projects_db) == 0
            changes = Change.query.filter_by(p_id=p_id).all()
            assert len(changes) == 0
            messages = Message.query.filter_by(p_id=p_id).all()
            assert len(messages) == 0



# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.test_user.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    tests for user related routes.

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2021 by the mss team, see AUTHORS.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import pytest
import requests
import json
import sys
import time

from PyQt5 import QtWidgets
from mslib.mscolab.server import check_login, register_user
from mslib.mscolab.conf import mscolab_settings
from mslib.msui.mscolab import MSSMscolabWindow
from mslib._tests.utils import mscolab_start_server


PORTS = list(range(9541, 9560))


@pytest.mark.skipif(os.name == "nt",
                    reason="multiprocessing needs currently start_method fork")
class Test_UserMethods(object):
    def setup(self):
        self.process, self.url, self.app, _, self.cm, self.fm = mscolab_start_server(PORTS)
        time.sleep(0.1)
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = MSSMscolabWindow(data_dir=mscolab_settings.MSCOLAB_DATA_DIR,
                                       mscolab_server_url=self.url)

    def teardown(self):
        if self.window.version_window:
            self.window.version_window.close()
        if self.window.conn:
            self.window.conn.disconnect()
        self.application.quit()
        QtWidgets.QApplication.processEvents()
        self.process.terminate()

    def test_registration(self):
        with self.app.app_context():
            x = register_user('sdf@s.com', 'sdf', 'sdf')
            assert x["success"] is True
            x = register_user('sdf@s.com', 'sdf', 'sdf')
            assert x["success"] is False
            x = register_user('sdf@ ssdf @s.com', 'sdf', 'sdf')
            assert x["success"] is False

    def test_login(self):
        with self.app.app_context():
            x = check_login('sdf@s.com', 'sdf')
            assert x is not None
            x = check_login('sdf@s.com', 'fd')
            assert x is not True

    def test_registration_api(self):
        data = {
            "email": "sdf@s1.com",
            "password": "sdf",
            "username": "sdf1"
        }
        r = requests.post(self.url + '/register', data=data).json()
        assert r["success"] is True
        r = requests.post(self.url + '/register', data=data).json()
        assert r["success"] is False

    def test_token_api(self):
        data = {
            "email": "sdf@s1.com",
            "password": "sdf",
            "username": "sdf1"
        }
        r = requests.post(self.url + '/register', data=data)
        r = requests.post(self.url + '/token', data=data)
        json_ = json.loads(r.text)
        assert json_.get("token", None) is not None
        data["password"] = "asdf"
        r = requests.post(self.url + '/token', data=data)
        assert r.text == "False"



# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.test_files_api
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    api integration tests for file based handlers

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import pytest
import requests
import json
import sys
import time

from PyQt5 import QtWidgets
from werkzeug.urls import url_join
from mslib.mscolab.models import User, Change, Project
from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.server import db
from mslib.mscolab.utils import get_recent_pid
from mslib._tests.utils import mscolab_register_and_login, mscolab_create_project, mscolab_start_server
from mslib.msui.mscolab import MSSMscolabWindow


PORTS = list(range(9381, 9400))


@pytest.mark.skipif(os.name == "nt",
                    reason="multiprocessing needs currently start_method fork")
class Test_Files(object):
    def setup(self):
        self.process, self.url, self.app, _, self.cm, self.fm = mscolab_start_server(PORTS)
        time.sleep(0.1)
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = MSSMscolabWindow(data_dir=mscolab_settings.MSCOLAB_DATA_DIR,
                                       mscolab_server_url=self.url)
        self.sockets = []
        self.file_message_counter = [0] * 2
        self.undefined_p_id = 123
        self.no_perm_p_id = 2
        data = {
            'email': 'a',
            'password': 'a'
        }
        r = requests.post(self.url + '/token', data=data)
        self.token = json.loads(r.text)['token']
        with self.app.app_context():
            self.user = User.query.filter_by(id=8).first()

    def teardown(self):
        for socket in self.sockets:
            socket.disconnect()
        if self.window.version_window:
            self.window.version_window.close()
        if self.window.conn:
            self.window.conn.disconnect()
        self.application.quit()
        QtWidgets.QApplication.processEvents()
        self.process.terminate()

    def test_create_project(self):
        data = {
            "token": self.token,
            "path": "dummy",
            "description": "test description"
        }
        url = url_join(self.url, 'create_project')
        r = requests.post(url, data=data)
        assert r.text == "True"
        r = requests.post(url, data=data)
        assert r.text == "False"

    def test_projects(self):
        data = {
            "token": self.token
        }
        url = url_join(self.url, 'projects')
        r = requests.get(url, data=data)
        json_res = json.loads(r.text)
        assert len(json_res["projects"]) == 3
        data["token"] = "garbage text"
        r = requests.get(url, data=data)
        assert r.text == "False"

    def test_get_project(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            data = {
                "token": self.token,
                "p_id": p_id
            }
            url = url_join(self.url, 'get_project_by_id')
            r = requests.get(url, data=data)
            assert json.loads(r.text)["content"] == self.fm.get_file(int(p_id), self.user)

    def test_authorized_users(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
        data = {
            "token": self.token,
            "p_id": p_id
        }
        url = url_join(self.url, 'authorized_users')
        r = requests.get(url, data=data)
        users = json.loads(r.text)["users"]
        assert len(users) == 2
        # for any other random process which doesn't exist it will return empty array
        data["p_id"] = 43
        r = requests.get(url, data=data)
        users = json.loads(r.text)["users"]
        assert len(users) == 0

    def test_get_users_without_permission(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
        data = {
            "token": self.token,
            "p_id": p_id
        }
        url = url_join(self.url, "users_without_permission")
        res = requests.get(url, data=data).json()
        assert res["success"] is True
        data["p_id"] = self.undefined_p_id
        res = requests.get(url, data=data).json()
        assert res["success"] is False
        data["p_id"] = self.no_perm_p_id
        res = requests.get(url, data=data).json()
        assert res["success"] is False

    def test_get_users_with_permission(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
        data = {
            "token": self.token,
            "p_id": p_id
        }
        url = url_join(self.url, "users_with_permission")
        res = requests.get(url, data=data).json()
        assert res["success"] is True
        data["p_id"] = self.undefined_p_id
        res = requests.get(url, data=data).json()
        assert res["success"] is False
        data["p_id"] = self.no_perm_p_id
        res = requests.get(url, data=data).json()
        assert res["success"] is False

    def test_add_bulk_permissions(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
        data = {
            "token": self.token,
            "p_id": p_id,
            "selected_userids": json.dumps([12, 13]),
            "selected_access_level": "collaborator"
        }
        url = url_join(self.url, 'add_bulk_permissions')
        res = requests.post(url, data=data).json()
        assert res["success"] is True
        data["p_id"] = self.undefined_p_id
        res = requests.post(url, data=data).json()
        assert res["success"] is False
        data["p_id"] = self.no_perm_p_id
        res = requests.post(url, data=data).json()
        assert res["success"] is False

    def test_modify_bulk_permissions(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            assert p_id == 4
        data = {
            "token": self.token,
            "p_id": p_id,
            "selected_userids": json.dumps([12, 13]),
            "selected_access_level": "viewer"
        }
        url = url_join(self.url, 'modify_bulk_permissions')
        r = requests.post(url, data=data).json()
        assert r["success"] is True
        data["p_id"] = self.undefined_p_id
        r = requests.post(url, data=data).json()
        assert r["success"] is False
        data["p_id"] = self.no_perm_p_id
        r = requests.post(url, data=data).json()
        assert r["success"] is False

    def test_delete_bulk_permissions(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
        data = {
            "token": self.token,
            "p_id": p_id,
            "selected_userids": json.dumps([12, 13]),
        }
        url = url_join(self.url, 'delete_bulk_permissions')
        r = requests.post(url, data=data).json()
        assert r["success"] is True
        data["p_id"] = self.undefined_p_id
        r = requests.post(url, data=data).json()
        assert r["success"] is False
        data["p_id"] = self.no_perm_p_id
        r = requests.post(url, data=data).json()
        assert r["success"] is False

    def test_import_permissions(self):
        current_p_id = 4
        import_p_id = 1
        data = {
            "token": self.token,
            "current_p_id": current_p_id,
            "import_p_id": import_p_id
        }
        url = url_join(self.url, 'import_permissions')
        res = requests.post(url, data=data).json()
        assert res["success"] is True
        data["import_p_id"] = self.no_perm_p_id
        res = requests.post(url, data=data).json()
        assert res["success"] is False
        data["current_p_id"] = self.no_perm_p_id
        res = requests.post(url, data=data).json()
        assert res["success"] is False

    def test_update_project(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
        # ToDo handle paths with blank characters here
        data = {
            "token": self.token,
            "p_id": p_id,
            "attribute": "path",
            "value": "a_diff_path"
        }
        get_proj_url = url_join(self.url, 'get_project_by_id')
        update_proj_url = url_join(self.url, 'update_project')
        r = requests.post(update_proj_url, data=data)
        assert r.text == "True"
        # to make sure that path has changed, which is indirectly known by this request
        # getting results properly
        r = requests.get(get_proj_url, data=data)
        assert r.text != "False"
        data = {
            "token": self.token,
            "p_id": p_id,
            "attribute": "description",
            "value": "a_diff description"
        }
        r = requests.post(update_proj_url, data=data)
        assert r.text == "True"
        with self.app.app_context():
            project = Project.query.filter_by(id=p_id).first()
            assert project.path == "a_diff_path"
            assert project.description == "a_diff description"

    def test_delete_project(self):
        with self.app.app_context():
            response = mscolab_register_and_login(self.app, self.url, 'a', 'a', 'a')
            assert response.status == '200 OK'
            data, response = mscolab_create_project(self.app, self.url, response,
                                                    path='f3', description='f3 test example')
            assert response.status == '200 OK'
            p_id = get_recent_pid(self.fm, self.user)

        # first free number after sseding
        assert p_id == 7
        data = {
            "token": self.token,
            "p_id": p_id
        }
        url = url_join(self.url, 'delete_project')
        res = requests.post(url, data=data).json()
        assert res["success"] is True

    def test_change(self):
        """
        since file needs to be saved to inflict changes, changes during integration
        tests have to be manually inserted
        """
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            ch = Change(int(p_id), 8, "", "Version1", "some comment")

            db.session.add(ch)
            db.session.commit()
        data = {
            "token": self.token,
            "p_id": p_id
        }
        # test 'get all changes' request
        url = url_join(self.url, 'get_all_changes')
        r = requests.get(url, data=data)
        changes = json.loads(r.text)["changes"]
        assert len(changes) == 1
        assert changes[0]["comment"] == "some comment"

    def test_change_content(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
            ch = Change(int(p_id), 8, "", "Version changed", "some comment")
            db.session.add(ch)
            db.session.commit()
            data = {
                "token": self.token,
                "p_id": p_id
            }
            get_proj_url = url_join(self.url, 'get_project_by_id')
            res = requests.get(get_proj_url, data=data)
            content = json.loads(res.text)["content"]
            change = Change.query.order_by(Change.created_at.desc()).first()
            data = {
                "token": self.token,
                "ch_id": change.id
            }
            get_change_content_url = url_join(self.url, 'get_change_content')
            res = requests.get(get_change_content_url, data=data).json()
            change_content = res["content"]
            assert content.strip() == change_content.strip()

    def test_set_version_name(self):
        with self.app.app_context():
            p_id = int(get_recent_pid(self.fm, self.user))
            ch = Change(int(p_id), 8, "", "Version changed", "some comment")
            db.session.add(ch)
            db.session.commit()
            change = Change.query.filter_by(p_id=p_id).order_by(Change.created_at.desc()).first()
        data = {
            "token": self.token,
            "ch_id": change.id,
            "p_id": p_id,
            "version_name": "Test Version Name"
        }
        url = url_join(self.url, 'set_version_name')
        res = requests.post(url, data=data).json()
        assert res["success"] is True
        with self.app.app_context():
            change = Change.query.filter_by(id=change.id).first()
            assert change.version_name == "Test Version Name"



# -*- coding: utf-8 -*-
"""

    mslib.retriever
    ~~~~~~~~~~~~~~~~~~~~

    automation within mss to create for instance a number of the same plots
    for several flights or several forecast steps

    This file is part of mss.

    :copyright: Copyright 2020 Joern Ungermann
    :copyright: Copyright 2020-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import sys
import argparse
import datetime
import io
import os
import xml
import requests
from fs import open_fs
import PIL.Image

import mslib
import mslib.utils
import mslib.msui
import mslib.msui.mpl_map
import mslib.msui.mss_qt
import mslib.thermolib

import matplotlib.pyplot as plt


TEXT_CONFIG = {
    "bbox": dict(boxstyle="round", facecolor="white", alpha=0.5, edgecolor="none"),
    "fontweight": "bold", "zorder": 4, "fontsize": 6, "clip_on": True}


def load_from_ftml(filename):
    """Load a flight track from an XML file at <filename>.
    """
    _dirname, _name = os.path.split(filename)
    _fs = open_fs(_dirname)
    datasource = _fs.open(_name)
    try:
        doc = xml.dom.minidom.parse(datasource)
    except xml.parsers.expat.ExpatError as ex:
        raise SyntaxError(str(ex))

    ft_el = doc.getElementsByTagName("FlightTrack")[0]

    waypoints_list = []
    for wp_el in ft_el.getElementsByTagName("Waypoint"):

        location = wp_el.getAttribute("location")
        lat = float(wp_el.getAttribute("lat"))
        lon = float(wp_el.getAttribute("lon"))
        flightlevel = float(wp_el.getAttribute("flightlevel"))
        comments = wp_el.getElementsByTagName("Comments")[0]
        # If num of comments is 0(null comment), then return ''
        if len(comments.childNodes):
            comments = comments.childNodes[0].data.strip()
        else:
            comments = ''

        waypoints_list.append((lat, lon, flightlevel, location, comments))
    return waypoints_list


def main():
    parser = argparse.ArgumentParser(description="""
       This script automatically retrieves and stores a set of plots for the
       configured flights. The configuration is placed within the normal
       MSS frontend JSON file. E.g.

       "automated_plotting": {
           "flights": [
               ["ST25", "01 SADPAP (stereo)", "500,50",
                "ST25-joern.ftml",
                "2019-07-01T00:00:00Z", "2019-09-01T12:00:00Z"]
           ],
           "hsecs": [
               ["https://mss-server/campaigns2019",
                "ecmwf.PVTropo01", "default", "4.0"],
               ["https://mss-server/campaigns2019",
                "ecmwf.ertel_potential_vorticity_pl", "ertel_potential_vorticity_bh", "200.0"]
           ],
           "vsecs": [
               ["https://mss-server/campaigns2019",
                "ecmwf.VS_ertel_potential_vorticity_ml", "ertel_potential_vorticity_bh"],
               ["https://mss-server/campaigns2019",
                "ecmwf.TroposphereInversionLayer", ""]
           ]
       }

       will plot flight "ST25" with configured map section "01 SADPAP (stereo)" and
       vertical range 500hPa to 50hPa from the given FTML file for init time
       "2019-07-01T00:00:00Z" and valid time "2019-09-01T12:00:00Z". The plots
       are defined in the hsecs (horizontal cross-sections) and vsecs (vertical
       cross-sections) entries given each the URL of the server, the layer name, the style,
       and, for hsec only, the elevation to plot (if necessary).
    """)
    parser.add_argument("-v", "--version", help="show version", action="store_true", default=False)
    parser.add_argument("--debug", help="show debugging log messages on console", action="store_true", default=False)
    parser.add_argument("--logfile", help="Specify logfile location. Set to empty string to disable.", action="store",
                        default=os.path.join(mslib.msui.constants.MSS_CONFIG_PATH, "mss_pyui.log"))
    args = parser.parse_args()

    if args.version:
        print("***********************************************************************")
        print("\n            Mission Support System (mss_retriever)\n")
        print("***********************************************************************")
        print("Documentation: http://mss.rtfd.io")
        print("Version:", mslib.__version__)
        sys.exit()

    mslib.utils.setup_logging(args)

    config = mslib.utils.config_loader()
    num_interpolation_points = config["num_interpolation_points"]
    num_labels = config["num_labels"]
    tick_index_step = num_interpolation_points // num_labels

    fig = plt.figure()
    for flight, section, vertical, filename, init_time, time in \
            config["automated_plotting"]["flights"]:
        params = mslib.utils.get_projection_params(
            config["predefined_map_sections"][section]["CRS"].lower())
        params["basemap"].update(config["predefined_map_sections"][section]["map"])
        wps = load_from_ftml(filename)
        wp_lats, wp_lons, wp_locs = [[x[i] for x in wps] for i in [0, 1, 3]]
        wp_presss = [mslib.thermolib.flightlevel2pressure(wp[2]) for wp in wps]
        for url, layer, style, elevation in config["automated_plotting"]["hsecs"]:
            fig.clear()
            ax = fig.add_subplot(111, zorder=99)
            bm = mslib.msui.mpl_map.MapCanvas(ax=ax, **(params["basemap"]))

            # plot path and labels
            bm.plot(wp_lons, wp_lats,
                    color="blue", marker="o", linewidth=2, markerfacecolor="red",
                    latlon=True, markersize=4, zorder=100)
            for i, (lon, lat, loc) in enumerate(zip(wp_lons, wp_lats, wp_locs)):
                textlabel = f"{loc if loc else str(i)}   "
                x, y = bm(lon, lat)
                plt.text(x, y, textlabel, **TEXT_CONFIG)
            plt.tight_layout()

            # retrieve and draw WMS image
            ax_bounds = plt.gca().bbox.bounds
            width, height = int(round(ax_bounds[2])), int(round(ax_bounds[3]))
            bbox = params['basemap']
            req = requests.get(
                url, auth=tuple(config["WMS_login"][url]),
                params={"version": "1.3.0", "request": "GetMap", "format": "image/png",
                        "exceptions": "XML",
                        "crs": config["predefined_map_sections"][section]["CRS"],
                        "layers": layer, "styles": style, "elevation": elevation,
                        "dim_init_time": init_time, "time": time,
                        "width": width, "height": height,
                        "bbox": f"{bbox['llcrnrlat']},{bbox['llcrnrlon']},{bbox['urcrnrlat']},{bbox['urcrnrlon']}"})
            if req.headers['Content-Type'] == "text/xml":
                print(flight, section, vertical, filename, init_time, time)
                print(url, layer, style, elevation)
                print("WMS Error:")
                print(req.text)
                exit(1)
            image_io = io.BytesIO(req.content)
            img = PIL.Image.open(image_io)
            bm.imshow(img, interpolation="nearest", origin="upper")
            bm.drawcoastlines()
            bm.drawcountries()

            fig.savefig(f"{flight}_{layer}.png")

        # prepare vsec plots
        path = [(wp[0], wp[1], datetime.datetime.now()) for wp in wps]
        lats, lons, _ = mslib.utils.path_points(
            path, numpoints=num_interpolation_points + 1, connection="greatcircle")
        intermediate_indexes = []
        ipoint = 0
        for i, (lat, lon) in enumerate(zip(lats, lons)):
            if abs(lat - wps[ipoint][0]) < 1E-10 and abs(lon - wps[ipoint][1]) < 1E-10:
                intermediate_indexes.append(i)
                ipoint += 1
            if ipoint >= len(wps):
                break

        for url, layer, style in config["automated_plotting"]["vsecs"]:
            fig.clear()

            # setup ticks and labels
            ax = fig.add_subplot(111, zorder=99)
            ax.set_yscale("log")
            p_bot, p_top = [float(x) * 100 for x in vertical.split(",")]
            bbox = ",".join(str(x) for x in (num_interpolation_points, p_bot / 100, num_labels, p_top / 100))
            ax.grid(b=True)
            ax.patch.set_facecolor("None")
            pres_maj = mslib.msui.mpl_qtwidget.MplSideViewCanvas._pres_maj
            pres_min = mslib.msui.mpl_qtwidget.MplSideViewCanvas._pres_min
            major_ticks = pres_maj[(pres_maj <= p_bot) & (pres_maj >= p_top)]
            minor_ticks = pres_min[(pres_min <= p_bot) & (pres_min >= p_top)]
            labels = [f"{int(_mt / 100)}"
                      if (_mt / 100.) - int(_mt / 100.) == 0 else f"{float(_mt / 100)}" for _mt in major_ticks]
            if len(labels) > 20:
                labels = ["" if _x.split(".")[-1][0] in "975" else _x for _x in labels]
            elif len(labels) > 10:
                labels = ["" if _x.split(".")[-1][0] in "9" else _x for _x in labels]
            ax.set_ylabel("pressure (hPa)")
            ax.set_yticks(minor_ticks, minor=True)
            ax.set_yticks(major_ticks, minor=False)
            ax.set_yticklabels([], minor=True, fontsize=10)
            ax.set_yticklabels(labels, minor=False, fontsize=10)
            ax.set_ylim(p_bot, p_top)
            ax.set_xlim(0, num_interpolation_points)
            ax.set_xticks(range(0, num_interpolation_points, tick_index_step))
            ax.set_xticklabels(
                [f"{x[0]:2.1f}, {x[1]:2.1f}"
                 for x in zip(lats[::tick_index_step], lons[::tick_index_step])],
                rotation=25, fontsize=10, horizontalalignment="right")
            ax.set_xlabel("lat/lon")

            # plot path and waypoint labels
            ax.plot(intermediate_indexes, wp_presss,
                    color="blue", marker="o", linewidth=2, markerfacecolor="red",
                    markersize=4)
            for i, (idx, press, loc) in enumerate(zip(intermediate_indexes, wp_presss, wp_locs)):
                textlabel = f"{loc if loc else str(i)} "
                plt.text(idx + 1, press, textlabel, rotation=90, **TEXT_CONFIG)
            plt.tight_layout()

            # retrieve and draw WMS image
            ax_bounds = plt.gca().bbox.bounds
            width, height = int(round(ax_bounds[2])), int(round(ax_bounds[3]))
            req = requests.get(
                url, auth=tuple(config["WMS_login"][url]),
                params={"version": "1.3.0", "request": "GetMap", "format": "image/png",
                        "exceptions": "XML",
                        "crs": "VERT:LOGP", "layers": layer, "styles": style,
                        "dim_init_time": init_time, "time": time,
                        "width": width, "height": height,
                        "path": ",".join(f"{wp[0]:.2f},{wp[1]:.2f}" for wp in wps),
                        "bbox": bbox})

            if req.headers['Content-Type'] == "text/xml":
                print(flight, section, vertical, filename, init_time, time)
                print(url, layer, style)
                print("WMS Error:")
                print(req.text)
                exit(1)
            image_io = io.BytesIO(req.content)
            img = PIL.Image.open(image_io)
            imgax = fig.add_axes(ax.get_position(), frameon=True,
                                 xticks=[], yticks=[], label="ax2", zorder=0)
            imgax.imshow(img, interpolation="nearest", aspect="auto", origin="upper")
            imgax.set_xlim(0, img.size[0] - 1)
            imgax.set_ylim(img.size[1] - 1, 0)

            plt.savefig(f"{flight}_{layer}.png")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.test_utils
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    tests for mscolab/utils

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2021 by the mss team, see AUTHORS.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import sys
import time
import pytest

from PyQt5 import QtWidgets

from mslib.mscolab.models import User
from mslib.mscolab.utils import get_recent_pid
from mslib.mscolab.conf import mscolab_settings
from mslib.msui.mscolab import MSSMscolabWindow
from mslib._tests.utils import mscolab_start_server


PORTS = list(range(9561, 9580))


@pytest.mark.skipif(os.name == "nt",
                    reason="multiprocessing needs currently start_method fork")
class Test_Utils(object):
    def setup(self):
        self.process, self.url, self.app, _, self.cm, self.fm = mscolab_start_server(PORTS)
        time.sleep(0.1)
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = MSSMscolabWindow(data_dir=mscolab_settings.MSCOLAB_DATA_DIR,
                                       mscolab_server_url=self.url)
        with self.app.app_context():
            self.user = User.query.filter_by(id=8).first()

    def teardown(self):
        if self.window.version_window:
            self.window.version_window.close()
        if self.window.conn:
            self.window.conn.disconnect()
        self.application.quit()
        QtWidgets.QApplication.processEvents()
        self.process.terminate()

    def test_get_recent_pid(self):
        with self.app.app_context():
            p_id = get_recent_pid(self.fm, self.user)
        assert p_id == 4



# -*- coding: utf-8 -*-
"""

    mslib.thermolib
    ~~~~~~~~~~~~~~~~

    Collection of thermodynamic functions.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# The function sat_vapour_pressure() has been ported from the IDL function
# 'VaporPressure' by Holger Voemel, available at http://cires.colorado.edu/~voemel/vp.html.

import numpy
import scipy.integrate
import logging


class VapourPressureError(Exception):
    """Exception class to handle error arising during the computation of vapour
       pressures.
    """

    def __init__(self, error_string):
        logging.debug("%s", error_string)


def sat_vapour_pressure(t, liquid='HylandWexler', ice='GoffGratch',
                        force_phase='None'):
    """
    Compute the saturation vapour pressure over liquid water and over ice
    with a variety of formulations.

    This function is a direct port from the IDL function 'VaporPressure' by
    Holger Voemel, available at http://cires.colorado.edu/~voemel/vp.html.

    By default, for temperatures > 0 degC, the saturation pressure over
    liquid water is computed; from temperatures <= 0 degC a formulation
    over ice is used.

    The current default fomulas are Hyland and Wexler for liquid and
    Goff Gratch for ice. (hv20040521)

    Arguments:
    t -- Temperature in [K]. Can be a scaler or an n-dimensional NumPy array.
    liquid -- Optional; specify the formulation for computing the saturation
              pressure over liquid water. Can be one of:

              HylandWexler, GoffGratch, Wexler, MagnusTeten, Buck_original,
              Buck_manual, WMO_Goff, WMO2000, Sonntag, Bolton, [Fukuta (N/A)],
              IAPWS, MurphyKoop.

    ice -- Optional; specify the formulation for computing the saturation
           pressure over ice. Can be one of:

           MartiMauersberger, HylandWexler, GoffGratch, MagnusTeten,
           Buck_original, Buck_manual, WMO_Goff, Sonntag, MurphyKoop.

           Please have a look at the source code for further information
           about the formulations.

    force_phase -- Optional; force liquid or ice phase to avoid automatic
                   switching of formulations at 0 degC. Can be 'liquid'
                   or 'ice'.

    Returns:
    Saturation vapor pressure [Pa], in the same dimensions as the input.
    """

    # Make sure the input is a NumPy array.
    if numpy.isscalar(t):
        t = numpy.array([t])
        input_scalar = True
    else:
        t = numpy.array(t)
        input_scalar = False

    # Get indexes of input temperatures above and below freezing, to select
    # the appropriate method for each temperature.
    if force_phase == "ice":
        idx_ice = ()  # numpy.where(t is not None)
        idx_liq = None
    elif force_phase == "liquid":
        idx_liq = ()  # numpy.where(t is not None)
        idx_ice = None
    elif force_phase == "None":
        idx_ice = numpy.where(t <= 273.15)
        idx_liq = numpy.where(t > 273.15)
    else:
        raise VapourPressureError("Cannot recognize the force_phase "
                                  f"keyword: '{force_phase}' (valid are ice, liquid, None)")

    # Initialise output field.
    e_sat = numpy.zeros(numpy.shape(t))

    # =============================================================================
    #  Calculate saturation pressure over liquid water ----------------------------
    if not force_phase == 'ice':

        if liquid == 'MartiMauersberger':
            raise VapourPressureError("Marti and Mauersberger don't "
                                      "have a vapour pressure curve over liquid.")

        elif liquid == 'HylandWexler':
            # Source: Hyland, R. W. and A. Wexler, Formulations for the
            # Thermodynamic Properties of the saturated Phases of H2O
            # from 173.15K to 473.15K, ASHRAE Trans, 89(2A), 500-519, 1983.
            e_sat[idx_liq] = (numpy.exp((-0.58002206E4 / t[idx_liq]) +
                                        0.13914993E1 -
                                        0.48640239E-1 * t[idx_liq] +
                                        0.41764768E-4 * t[idx_liq] ** 2. -
                                        0.14452093E-7 * t[idx_liq] ** 3. +
                                        0.65459673E1 * numpy.log(t[idx_liq])) / 100.)

        elif liquid == 'Wexler':
            # Wexler, A., Vapor pressure formulation for ice, Journal of
            # Research of the National Bureau of Standards-A. 81A, 5-20, 1977.
            e_sat[idx_liq] = (numpy.exp(-2.9912729E3 * t[idx_liq] ** (-2.) -
                                        6.0170128E3 * t[idx_liq] ** (-1.) +
                                        1.887643854E1 * t[idx_liq] ** 0. -
                                        2.8354721E-2 * t[idx_liq] ** 1. +
                                        1.7838301E-5 * t[idx_liq] ** 2. -
                                        8.4150417E-10 * t[idx_liq] ** 3. -
                                        4.4412543E-13 * t[idx_liq] ** 4. +
                                        2.858487 * numpy.log(t[idx_liq])) / 100.)

        elif liquid == 'GoffGratch':
            # Goff Gratch formulation.
            # Source: Smithsonian Meteorological Tables, 5th edition,
            # p. 350, 1984
            # From original source: Goff and Gratch (1946), p. 107.
            ts = 373.16  # steam point temperature in K
            ews = 1013.246  # saturation pressure at steam point
            # temperature, normal atmosphere
            e_sat[idx_liq] = 10. ** (-7.90298 * ((ts / t[idx_liq]) - 1.) +
                                     5.02808 * numpy.log10((ts / t[idx_liq])) -
                                     1.3816E-7 * (10. ** (11.344 * (1. - (t[idx_liq] / ts))) - 1.) +
                                     8.1328E-3 * (10. ** (-3.49149 * ((ts / t[idx_liq]) - 1)) - 1.) +
                                     numpy.log10(ews))

        elif liquid == 'MagnusTeten':
            # Source: Murray, F. W., On the computation of saturation
            # vapor pressure, J. Appl. Meteorol., 6, 203-204, 1967.
            tc = t - 273.15
            e_sat[idx_liq] = 10. ** (7.5 * (tc[idx_liq]) / (tc[idx_liq] + 237.5) + 0.7858)

        elif liquid == 'Buck_original':
            # Bucks vapor pressure formulation based on Tetens formula
            # Source: Buck, A. L., New equations for computing vapor
            # pressure and enhancement factor, J. Appl. Meteorol., 20,
            # 1527-1532, 1981.
            tc = t - 273.15
            e_sat[idx_liq] = 6.1121 * numpy.exp(17.502 * tc[idx_liq] / (240.97 + tc[idx_liq]))

        elif liquid == 'Buck_manual':
            # Bucks vapor pressure formulation based on Tetens formula
            # Source: Buck Research, Model CR-1A Hygrometer Operating
            # Manual, Sep 2001
            tc = t - 273.15
            e_sat[idx_liq] = 6.1121 * numpy.exp((18.678 - (tc[idx_liq] / 234.5)) *
                                                (tc[idx_liq]) / (257.14 + tc[idx_liq]))

        elif liquid == 'WMO_Goff':
            # Intended WMO formulation, originally published by Goff (1957)
            # incorrectly referenced by WMO technical regulations, WMO-NO 49,
            # Vol I, General Meteorological Standards and Recommended
            # Practices, App. A, Corrigendum Aug 2000.
            # and incorrectly referenced by WMO technical regulations,
            # WMO-NO 49, Vol I, General Meteorological Standards and
            # Recommended Practices, App. A, 1988.
            ts = 273.16  # steam point temperature in K
            e_sat[idx_liq] = 10. ** (10.79574 * (1. - (ts / t[idx_liq])) -
                                     5.02800 * numpy.log10((t[idx_liq] / ts)) +
                                     1.50475E-4 * (1. - 10. ** (-8.2969 * ((t[idx_liq] / ts) - 1.))) +
                                     0.42873E-3 * (10. ** (+4.76955 * (1. - (ts / t[idx_liq]))) - 1.) +
                                     0.78614)

        elif liquid == 'WMO2000':
            # WMO formulation, which is very similar to Goff Gratch
            # Source: WMO technical regulations, WMO-NO 49, Vol I,
            # General Meteorological Standards and Recommended Practices,
            # App. A, Corrigendum Aug 2000.
            ts = 273.16  # steam point temperature in K
            e_sat[idx_liq] = 10. ** (10.79574 * (1. - (ts / t[idx_liq])) -
                                     5.02800 * numpy.log10((t[idx_liq] / ts)) +
                                     1.50475E-4 * (1. - 10. ** (-8.2969 * ((t[idx_liq] / ts) - 1.))) +
                                     0.42873E-3 * (10. ** (-4.76955 * (1. - (ts / t[idx_liq]))) - 1.) +
                                     0.78614)

        elif liquid == 'Sonntag':
            # Source: Sonntag, D., Advancements in the field of hygrometry,
            # Meteorol. Z., N. F., 3, 51-66, 1994.
            e_sat[idx_liq] = numpy.exp(-6096.9385 * t[idx_liq] ** (-1.) +
                                       16.635794 -
                                       2.711193E-2 * t[idx_liq] ** 1. +
                                       1.673952E-5 * t[idx_liq] ** 2. +
                                       2.433502 * numpy.log(t[idx_liq]))

        elif liquid == 'Bolton':
            # Source: Bolton, D., The computation of equivalent potential
            # temperature, Monthly Weather Report, 108, 1046-1053, 1980.
            # equation (10)
            tc = t - 273.15
            e_sat[idx_liq] = 6.112 * numpy.exp(17.67 * tc[idx_liq] / (tc[idx_liq] + 243.5))

        # THIS CURVE LOOKS WRONG!
        #         elif liquid == 'Fukuta':
        #             # Source: Fukuta, N. and C. M. Gramada, Vapor pressure
        #             # measurement of supercooled water, J. Atmos. Sci., 60,
        #             # 1871-1875, 2003.
        #             # This paper does not give a vapor pressure formulation,
        #             # but rather a correction over the Smithsonian Tables.
        #             # Thus calculate the table value first, then use the
        #             # correction to get to the measured value.
        #             ts    = 373.16       # steam point temperature in K
        #             ews   = 1013.246     # saturation pressure at steam point
        #                                  # temperature, normal atmosphere

        #             e_sat[idx_liq] = 10.**(-7.90298*(ts/t[idx_liq]-1.)
        #                                    + 5.02808 * numpy.log10(ts/t[idx_liq])
        #                                    - 1.3816E-7 * (10.**(11.344*(1.-t[idx_liq]/ts))-1.)
        #                                    + 8.1328E-3*(10.**(-3.49149*(ts/t[idx_liq]-1)) -1.)
        #                                    + numpy.log10(ews))

        #             tc = t - 273.15
        #             x = tc[idx_liq] + 19
        #             e_sat[idx_liq] = e_sat[idx_liq] * (0.9992 + 7.113E-4*x
        #                                                - 1.847E-4*x**2.
        #                                                + 1.189E-5*x**3.
        #                                                + 1.130E-7*x**4.
        #                                                - 1.743E-8*x**5.)

        #             e_sat[numpy.where(tc < -39.)] = None

        elif liquid == 'IAPWS':
            # Source: Wagner W. and A. Pruss (2002), The IAPWS
            # formulation 1995 for the thermodynamic properties
            # of ordinary water substance for general and scientific
            # use, J. Phys. Chem. Ref. Data, 31(2), 387-535.
            # This is the 'official' formulation from the International
            # Association for the Properties of Water and Steam
            # The valid range of this formulation is 273.16 <= T <=
            # 647.096 K and is based on the ITS90 temperature scale.
            Tc = 647.096  # K   : Temperature at the critical point
            Pc = 22.064 * 10 ** 4  # hPa : Vapor pressure at the critical point
            nu = (1. - (t[idx_liq] / Tc))
            a1 = -7.85951783
            a2 = 1.84408259
            a3 = -11.7866497
            a4 = 22.6807411
            a5 = -15.9618719
            a6 = 1.80122502
            e_sat[idx_liq] = Pc * numpy.exp(Tc / t[idx_liq] *
                                            (a1 * nu + a2 * nu ** 1.5 + a3 * nu ** 3. +
                                             a4 * nu ** 3.5 + a5 * nu ** 4. + a6 * nu ** 7.5))

        elif liquid == 'MurphyKoop':
            # Source : Murphy and Koop, Review of the vapour pressure
            # of ice and supercooled water for atmospheric applications,
            # Q. J. R. Meteorol. Soc (2005), 131, pp. 1539-1565.
            e_sat[idx_liq] = (numpy.exp(54.842763 - (6763.22 / t[idx_liq]) -
                                        4.210 * numpy.log(t[idx_liq]) +
                                        0.000367 * t[idx_liq] +
                                        numpy.tanh(0.0415 * (t[idx_liq] - 218.8)) *
                                        (53.878 - (1331.22 / t[idx_liq]) -
                                         9.44523 * numpy.log(t[idx_liq]) +
                                         0.014025 * t[idx_liq])) / 100.)

        else:
            raise VapourPressureError("Unkown method for computing "
                                      f"the vapour pressure curve over liquid: {liquid}")

    # =============================================================================
    #  Calculate saturation pressure over ice -------------------------------------
    if not force_phase == 'liquid':

        if ice == 'WMO2000':
            ice = 'WMO_Goff'

        if ice == 'IAWPS':
            raise VapourPressureError("IAPWS does not provide a vapour "
                                      "pressure formulation over ice")

        elif ice == 'MartiMauersberger':
            # Source: Marti, J. and K Mauersberger, A survey and new
            # measurements of ice vapor pressure at temperatures between
            # 170 and 250 K, GRL 20, 363-366, 1993.
            e_sat[idx_ice] = (10. ** ((-2663.5 / t[idx_ice]) + 12.537) / 100.)

        elif ice == 'HylandWexler':
            # Source Hyland, R. W. and A. Wexler, Formulations for the
            # Thermodynamic Properties of the saturated Phases of H2O
            # from 173.15K to 473.15K, ASHRAE Trans, 89(2A), 500-519, 1983.
            e_sat[idx_ice] = (numpy.exp((-0.56745359E4 / t[idx_ice]) +
                                        0.63925247E1 -
                                        0.96778430E-2 * t[idx_ice] +
                                        0.62215701E-6 * t[idx_ice] ** 2. +
                                        0.20747825E-8 * t[idx_ice] ** 3. -
                                        0.94840240E-12 * t[idx_ice] ** 4. +
                                        0.41635019E1 * numpy.log(t[idx_ice])) / 100.)

        elif ice == 'GoffGratch':
            # Source: Smithsonian Meteorological Tables, 5th edition,
            # p. 350, 1984

            ei0 = 6.1071  # mbar
            T0 = 273.16  # freezing point in K

            e_sat[idx_ice] = 10. ** (-9.09718 * ((T0 / t[idx_ice]) - 1.) -
                                     3.56654 * numpy.log10((T0 / t[idx_ice])) +
                                     0.876793 * (1. - (t[idx_ice] / T0)) +
                                     numpy.log10(ei0))

        elif ice == 'MagnusTeten':
            # Source: Murray, F. W., On the computation of saturation
            # vapour pressure, J. Appl. Meteorol., 6, 203-204, 1967.
            tc = t - 273.15
            e_sat[idx_ice] = 10. ** (9.5 * tc[idx_ice] / (265.5 + tc[idx_ice]) + 0.7858)

        elif ice == 'Buck_original':
            # Bucks vapor pressure formulation based on Tetens formula
            # Source: Buck, A. L., New equations for computing vapor
            # pressure and enhancement factor, J. Appl. Meteorol., 20,
            # 1527-1532, 1981.
            tc = t - 273.15
            e_sat[idx_ice] = 6.1115 * numpy.exp(22.452 * tc[idx_ice] / (272.55 + tc[idx_ice]))

        elif ice == 'Buck_manual':
            # Bucks vapor pressure formulation based on Tetens formula
            # Source: Buck Research, Model CR-1A Hygrometer Operating
            # Manual, Sep 2001
            tc = t - 273.15
            e_sat[idx_ice] = 6.1115 * numpy.exp((23.036 - (tc[idx_ice] / 333.7)) *
                                                tc[idx_ice] / (279.82 + tc[idx_ice]))

        elif ice == 'WMO_Goff':
            # WMO formulation, which is very similar to Goff Gratch
            # Source: WMO technical regulations, WMO-NO 49, Vol I,
            # General Meteorological Standards and Recommended Practices,
            # Aug 2000, App. A.

            T0 = 273.16  # steam point temperature in K

            e_sat[idx_ice] = 10. ** (-9.09685 * ((T0 / t[idx_ice]) - 1.) -
                                     3.56654 * numpy.log10((T0 / t[idx_ice])) +
                                     0.87682 * (1. - (t[idx_ice] / T0)) + 0.78614)

        elif ice == 'Sonntag':
            # Source: Sonntag, D., Advancements in the field of hygrometry,
            # Meteorol. Z., N. F., 3, 51-66, 1994.
            e_sat[idx_ice] = numpy.exp(-6024.5282 * t[idx_ice] ** (-1.) +
                                       24.721994 +
                                       1.0613868E-2 * t[idx_ice] ** 1. -
                                       1.3198825E-5 * t[idx_ice] ** 2. -
                                       0.49382577 * numpy.log(t[idx_ice]))

        elif ice == 'MurphyKoop':
            # Source: Murphy and Koop, Review of the vapour pressure of ice
            # and supercooled water for atmospheric applications, Q. J. R.
            # Meteorol. Soc (2005), 131, pp. 1539-1565.
            e_sat[idx_ice] = (numpy.exp(9.550426 - (5723.265 / t[idx_ice]) +
                                        3.53068 * numpy.log(t[idx_ice]) -
                                        0.00728332 * t[idx_ice]) / 100.)

        else:
            raise VapourPressureError("Unkown method for computing "
                                      f"the vapour pressure curve over ice: {ice}")

    # Convert return value units from hPa to Pa.
    return e_sat * 100. if not input_scalar else e_sat[0] * 100.


def rel_hum(p, t, q, liquid='HylandWexler', ice='GoffGratch',
            force_phase='None'):
    """Compute relative humidity in [%] from pressure, temperature, and
       specific humidity.

    Arguments:
    p -- pressure in [Pa]
    t -- temperature in [K]
    q -- specific humidity in [kg/kg]

    p, t and q can be scalars of NumPy arrays. They just have to either all
    scalars, or all arrays.

    liquid, ice, force_phase -- optional keywords to control the calculation
                                of the saturation vapour pressure; see
                                help of function 'sat_vapour_pressure()' for
                                details.

    Returns: Relative humidity in [%]. Same dimension as input fields.
    """
    if not (numpy.isscalar(p) or numpy.isscalar(t) or numpy.isscalar(q)):
        if not isinstance(p, numpy.ndarray):
            p = numpy.array(p)
        if not isinstance(t, numpy.ndarray):
            t = numpy.array(t)
        if not isinstance(q, numpy.ndarray):
            q = numpy.array(q)

    # Compute mixing ratio w from specific humidiy q.
    w = q / (1. - q)

    # Compute saturation vapour pressure from temperature t.
    e_sat = sat_vapour_pressure(t, liquid=liquid, ice=ice,
                                force_phase=force_phase)

    # Compute saturation mixing ratio from e_sat and pressure p.
    w_sat = 0.622 * e_sat / (p - e_sat)

    # Return the relative humidity, computed from w and w_sat.
    return 100. * w / w_sat


def virt_temp(t, q, method='exact'):
    """
    Compute virtual temperature in [K] from temperature and
    specific humidity.

    Arguments:
    t -- temperature in [K]
    q -- specific humidity in [kg/kg]

    t and q can be scalars of NumPy arrays. They just have to either all
    scalars, or all arrays.

    method -- optional keyword to specify the equation used. Default is
              'exact', which uses
                      Tv = T * (q + 0.622(1-q)) / 0.622,
              'approx' uses
                      Tv = T * (1 + 0.61w),
              with w = q/(1-q) being the water vapour mixing ratio.

              Reference: Wallace&Hobbs 2nd ed., eq. 3.16, 3.59, and 3.60
              (substitute w=q/(1-q) in 3.16 and 3.59 to obtain the exact
              formula).

    Returns: Virtual temperature in [K]. Same dimension as input fields.
    """
    if not (numpy.isscalar(t) or numpy.isscalar(q)):
        if not isinstance(t, numpy.ndarray):
            t = numpy.array(t)
        if not isinstance(q, numpy.ndarray):
            q = numpy.array(q)

    if method == 'exact':
        return t * (q + 0.622 * (1. - q)) / 0.622
    elif method == 'approx':
        # Compute mixing ratio w from specific humidiy q.
        w = q / (1. - q)
        return t * (1. + 0.61 * w)
    else:
        raise TypeError('virtual temperature method not understood')


def geop_difference(p, t, method='trapz', axis=-1):
    """Compute geopotential difference in [m**2 s**-2] between the pressure
       levels given by the first and last element in p (= pressure).

    Implements the hypsometric equation (1.17) from Holton, 3rd edition (or
    alternatively the integral form of (3.23) in Wallace and Hobbs, 2nd ed.).

    Arguments:
    p -- pressure in [Pa], needs to be a NumPy array with at least 2 elements.
    t -- temperature in [K], needs to be a NumPy array with at least 2 elements.

         Both arrays can be multidimensional, in this case pay attention to
         the 'axis' argument.

    method -- optional keyword to specify the integration method used.
              Default is 'trapz', which uses the trapezoidal rule.
              Alternatively, 'simps' causes Simpson's rule to be used.
              'cumtrapz' returns an array with the integrals between the
              first value in p and all other values. This is useful, for
              instance, for computing the geopotential on all model
              levels.

              See the 'scipy.integrate' documentation for further details.

    axis -- optional keyword to specify the vertical coordinate axis if p, t
            are multidimensional (e.g. if the axes of p, t specify [time,
            level, lat, lon] set axis=1). Default is the last dimension.

    Returns: Geopotential difference between p[0] and p[-1] in [m**2 s**-2].
             If 'cumtrapz' is specified, an array of dimension dim(p)-1
             will be returned, in which value n represents the geopotential
             difference between p[0] and p[n+1].
    """

    # The hypsometric equation integrates over ln(p).
    lnp = numpy.log(p)

    # Use scipy.intgerate to evaluate the integral. It is
    #     phi2 - phi1 = Rd * int( T,  d ln(p), p1, p2 ),
    # where phi denotes the geopotential.
    if method == 'trapz':
        return 287.058 * scipy.integrate.trapz(t, lnp, axis=axis)
    elif method == 'cumtrapz':
        return 287.058 * scipy.integrate.cumtrapz(t, lnp, axis=axis)
    elif method == 'simps':
        return 287.058 * scipy.integrate.simps(t, lnp, axis=axis)
    else:
        raise TypeError('integration method for geopotential not understood')


def geop_thickness(p, t, q=None, cumulative=False, axis=-1):
    """
    Compute the geopotential thickness in [m] between the pressure levels
    given by the first and last element in p (= pressure).

    Implements the hypsometric equation (1.18) from Holton, 3rd edition (or
    alternatively (3.24) in Wallace and Hobbs, 2nd ed.).

    Arguments:
    p -- pressure in [Pa]
    t -- temperature in [K]
    q -- [optional] specific humidity in [kg/kg]. If q is given, T will
         be converted to virtual temperature to account for the effects
         of moisture in the air.

    All inputs need to be NumPy arrays with at least 2 elements.

    cumulative -- optional keyword to specify whether the single geopotential
                  thickness between p[0] and p[-1] is returned (False, default),
                  or whether an array containing the thicknesses between
                  p[0] and all other elements in p is returned (True). The
                  latter option is useful for computing the geopotential height
                  of all model levels.

    axis -- see geop_difference().

    Uses geop_difference() for the integral in the above equations.

    Returns: Geopotential thickness between p[0] and p[-1] in [m].
             If 'cumtrapz' is specified, an array of dimension dim(p)-1
             will be returned, in which value n represents the geopotential
             thickness between p[0] and p[n+1].
    """

    # Check whether humidity effects should be considered. If q is specified,
    # simply evaluate the hypsometric equation with virtual temperature instead
    # of absolute temperature (see Wallace and Hobbs, 2nd ed., section 3.2.1).
    if q is None:
        tv = t
    else:
        tv = virt_temp(t, q)

    # Evaluate equation 3.24 in Wallace and Hobbs:
    #     delta Z = -Rd/g0 * int( Tv,  d ln(p), p1, p2 ),
    # where Z denotes the geopotential height, Z = phi/g0.
    return -1. / 9.80665 * geop_difference(p, tv, method='cumtrapz' if cumulative else 'trapz', axis=axis)


def spec_hum_from_pTd(p, td, liquid='HylandWexler'):
    """
    Computes specific humidity in [kg/kg] from pressure and dew point
    temperature.

    Arguments:
    p -- pressure in [Pa]
    td -- dew point temperature in [K]

    p and td can be scalars or NumPy arrays. They just have to either both
    scalars, or both arrays.

    liquid -- optional keyword to specify the method used for computing the
              saturation water wapour. See sat_vapour_pressure() for
              further details.

    Returns: specific humidity in [kg/kg]. Same dimensions as the inputs.

    Method:
      Specific humidity q = w / (1+w), with w = mixing ratio. (Wallace & Hobbs,
      2nd ed., (3.57)). W&H write: 'The dew point [Td] is the temperature at
      which the saturation mixing ratio ws with respect to liquid water becomes
      equal to the actual mixing ratio w.'. Hence we need ws(Td).
      From W&H 3.62, we get ws = 0.622 * es / (p-es). Plugging this into the
      above equation for q and simplifying, we get
                      q = 0.622 * es / (p + es * [0.622-1.])
    """
    # Compute saturation vapour pressure from dew point temperature td.
    e_sat = sat_vapour_pressure(td, liquid=liquid)

    return 0.622 * e_sat / (p + e_sat * (0.622 - 1.))


def dewpoint_approx(p, q, method='Bolton'):
    """
    Computes dew point in [K] from pressure and specific humidity.

    Arguments:
    p -- pressure in [Pa]
    q -- specific humidity in [kg/kg]

    p and q can be scalars or NumPy arrays. They just have to either both
    scalars, or both arrays.

    method -- optional keyword to specify the method used to approximate
              the dew point temperature. Valid values are:

              'Bolton' (default): Use the inversion of Bolton (1980), eq.
              10, to compute dewpoint. According to Bolton, this is accurate
              to 0.03 K in the range 238..308 K. See also Emanuel (1994,
              'Atmospheric Convection', eq. 4.6.2).

    Returns: dew point temperature in [K].
    """
    if not (numpy.isscalar(p) or numpy.isscalar(q)):
        if not isinstance(p, numpy.ndarray):
            p = numpy.array(p)
        if not isinstance(q, numpy.ndarray):
            q = numpy.array(q)

    # Compute mixing ratio w from specific humidiy q.
    w = q / (1. - q)

    # Compute vapour pressure from pressure and mixing ratio
    # (Wallace and Hobbs 2nd ed. eq. 3.59).
    e_q = w / (w + 0.622) * p

    if method == 'Bolton':
        td = (243.5 / ((17.67 / numpy.log(e_q / 100. / 6.112)) - 1)) + 273.15
    else:
        raise ValueError(f"invalid dew point method '{method}'")

    return td


def pot_temp(p, t):
    """
    Computes potential temperature in [K] from pressure and temperature.

    Arguments:
    p -- pressure in [Pa]
    t -- temperature in [K]

    p and t can be scalars of NumPy arrays. They just have to either both
    scalars, or both arrays.

    Returns: potential temperature in [K]. Same dimensions as the inputs.

    Method:
                            theta = T * (p0/p)^(R/cp)
      with p0 = 100000. Pa, R = 287.058 JK-1kg-1, cp = 1004 JK-1kg-1.
    """
    return t * (100000. / p) ** (287.058 / 1004.)


def eqpt_approx(p, t, q, liquid='HylandWexler', ice='GoffGratch',
                force_phase='None'):
    """
    Computes equivalent potential temperature in [K] from pressure,
    temperature and specific humidity.

    Arguments:
    p -- pressure in [Pa]
    t -- temperature in [K]
    q -- specific humidity in [kg/kg]

    p, t and q can be scalars or NumPy arrays.

    Returns: equivalent potential temperature in [K]. Same dimensions as
    the inputs.

    Method:
                  theta_e = theta * exp((Lv*w_sat)/(cp*T))
      with theta = potential temperature (see pot_temp()), Lv = 2.25e6 Jkg-1,
      cp = 1004 JK-1kg-1.

    Reference:  Wallace & Hobbs, 2nd ed., eq. 3.71
    """
    # Compute potential temperature from p and t.
    theta = pot_temp(p, t)

    # Compute saturation vapour pressure from temperature t.
    e_sat = sat_vapour_pressure(t, liquid=liquid, ice=ice,
                                force_phase=force_phase)

    # Compute saturation mixing ratio from e_sat and pressure p.
    w_sat = 0.622 * e_sat / (p - e_sat)

    # Latent heat of evaporation.
    Lv = 2.25 * 1.e6
    cp = 1004.

    # Equation 3.71 from Wallace & Hobbs, 2nd ed.
    theta_e = theta * numpy.exp((Lv * w_sat) / (cp * t))
    return theta_e


def omega_to_w(omega, p, t):
    """
    Convert pressure vertical velocity to geometric vertical velocity.

    Arguments:
    omega -- vertical velocity in pressure coordinates, in [Pa/s]
    p -- pressure in [Pa]
    t -- temperature in [K]

    All inputs can be scalars or NumPy arrays.

    Returns the vertical velocity in geometric coordinates, [m/s].

    For all grid points, the pressure vertical velocity in Pa/s is converted
    to m/s via
                    w[m/s] =(approx) omega[Pa/s] / (-g*rho)
                       rho = p / R*T
    with R = 287.058 JK-1kg-1, g = 9.80665 m2s-2.
    (see p.13 of 'Introduction to circulating atmospheres' by Ian N. James).

    NOTE: Please check the resulting values, especially in the upper atmosphere!
    """
    rho = p / (287.058 * t)
    return (omega / (-9.80665 * rho))


def flightlevel2pressure(flightlevel):
    """Conversion of flight level (given in hft) to pressure (Pa) with
       hydrostatic equation, according to the profile of the ICAO
       standard atmosphere.

    Reference:
        For example, H. Kraus, Die Atmosphaere der Erde, Springer, 2001,
        470pp., Sections II.1.4. and II.6.1.2.

    Arguments:
        flightlevel -- flight level in hft
    Returns:
        static pressure (Pa)
    """
    # Convert flight level (ft) to m (1 ft = 30.48 cm; 1/0.3048m = 3.28...).
    z = flightlevel * 30.48

    # g and R are used by all equations below.
    g = 9.80665
    R = 287.058

    if z <= 11000.:
        # ICAO standard atmosphere between 0 and 11 km: T(z=0km) = 15 degC,
        # p(z=0km) = 1013.25 hPa. Temperature gradient is 6.5 K/km.
        z0 = 0.
        T0 = 288.15
        gamma = 6.5e-3
        p0 = 101325.

        # Hydrostatic equation with linear temperature gradient.
        p = p0 * ((T0 - gamma * z - z0) / T0) ** (g / (gamma * R))
        return p

    elif z <= 20000.:
        # ICAO standard atmosphere between 11 and 20 km: T(z=11km) = -56.5 degC,
        # p(z=11km) = 226.32 hPa. Temperature is constant at -56.5 degC.
        z0 = 11000.
        p0 = 22632.64
        T = 216.65

        # Hydrostatic equation with constant temperature profile.
        p = p0 * numpy.exp(-g * (z - z0) / (R * T))
        return p

    elif z <= 32000.:
        # ICAO standard atmosphere between 20 and 32 km: T(z=20km) = -56.5 degC,
        # p(z=20km) = 54.75 hPa. Temperature gradient is -1.0 K/km.
        z0 = 20000.
        T0 = 216.65
        gamma = -1.0e-3
        p0 = 5475.16

        # Hydrostatic equation with linear temperature gradient.
        p = p0 * ((T0 - gamma * (z - z0)) / T0) ** (g / (gamma * R))
        return p

    elif z <= 47000.:
        # ICAO standard atmosphere between 32 and 47 km: T(z=32km) = -44.5 degC,
        # p(z=32km) = 8.68019 hPa. Temperature gradient is -2.8 K/km.
        z0 = 32000.
        T0 = 228.66
        gamma = -2.8e-3
        p0 = 868.089

        # Hydrostatic equation with linear temperature gradient.
        p = p0 * ((T0 - gamma * (z - z0)) / T0) ** (g / (gamma * R))
        return p

    elif z <= 51000:
        # ICAO standard atmosphere between 47 and 51 km: T(z=47km) = -2.5 degC,
        # p(z=47km) = 1.10906 hPa. Temperature is constant at -2.5 degC.
        z0 = 47000.
        p0 = 110.928
        T = 270.65

        # Hydrostatic equation with constant temperature profile.
        p = p0 * numpy.exp(-g * (z - z0) / (R * T))
        return p

    elif z <= 71000:
        # ICAO standard atmosphere between 51 and 71 km: T(z=51km) = -2.5 degC,
        # p(z=71km) = 0.66939 hPa. Temperature gradient is 2.8 K/km.
        z0 = 51000.
        T0 = 270.65
        gamma = 2.8e-3
        p0 = 66.952

        # Hydrostatic equation with linear temperature gradient.
        p = p0 * ((T0 - gamma * (z - z0)) / T0) ** (g / (gamma * R))
        return p

    else:
        raise ValueError("flight level to pressure conversion not "
                         "implemented for z > 71km")


def pressure2flightlevel(p):
    """Conversion of pressure (Pa) to flight level (hft) with
       hydrostatic equation, according to the profile of the ICAO
       standard atmosphere.

    Reference:
        For example, H. Kraus, Die Atmosphaere der Erde, Springer, 2001,
        470pp., Sections II.1.4. and II.6.1.2.

    Arguments:
        p -- pressure (Pa)
    Returns:
        flight level in hft
    """
    # g and R are used by all equations below.
    g = 9.80665
    R = 287.058

    if p < 3.956:
        raise ValueError("pressure to flight level conversion not "
                         "implemented for z > 71km (p ~ 4 Pa)")

    elif p <= 66.952:
        # ICAO standard atmosphere between 51 and 71 km: T(z=51km) = -2.5 degC,
        # p(z=71km) = 0.66939 hPa. Temperature gradient is 2.8 K/km.
        z0 = 51000.
        T0 = 270.65
        gamma = 2.8e-3
        p0 = 66.952

        # Hydrostatic equation with linear temperature gradient.
        z = z0 + 1. / gamma * (T0 - T0 * numpy.exp(gamma * R / g * numpy.log(p / p0)))

    elif p < 110.928:
        # ICAO standard atmosphere between 47 and 51 km: T(z=47km) = -2.5 degC,
        # p(z=47km) = 1.10906 hPa. Temperature is constant at -2.5 degC.
        z0 = 47000.
        p0 = 110.928
        T = 270.65

        # Hydrostatic equation with constant temperature profile.
        z = z0 - (R * T) / g * numpy.log(p / p0)

    elif p < 868.089:
        # ICAO standard atmosphere between 32 and 47 km: T(z=32km) = -44.5 degC,
        # p(z=32km) = 54.75 hPa. Temperature gradient is -2.8 K/km.
        z0 = 32000.
        T0 = 228.66
        gamma = -2.8e-3
        p0 = 868.089

        # Hydrostatic equation with linear temperature gradient.
        z = z0 + 1. / gamma * (T0 - T0 * numpy.exp(gamma * R / g * numpy.log(p / p0)))

    elif p < 5474.16:
        # ICAO standard atmosphere between 20 and 32 km: T(z=20km) = -56.5 degC,
        # p(z=20km) = 54.75 hPa. Temperature gradient is -1.0 K/km.
        z0 = 20000.
        T0 = 216.65
        gamma = -1.0e-3
        p0 = 5475.16

        # Hydrostatic equation with linear temperature gradient.
        z = z0 + 1. / gamma * (T0 - T0 * numpy.exp(gamma * R / g * numpy.log(p / p0)))

    elif p < 22632.:
        # ICAO standard atmosphere between 11 and 20 km: T(z=11km) = -56.5 degC,
        # p(z=11km) = 226.32 hPa. Temperature is constant at -56.5 degC.
        z0 = 11000.
        p0 = 22632.64
        T = 216.65

        # Hydrostatic equation with constant temperature profile.
        z = z0 - (R * T) / g * numpy.log(p / p0)

    else:
        # ICAO standard atmosphere between 0 and 11 km: T(z=0km) = 15 degC,
        # p(z=0km) = 1013.25 hPa. Temperature gradient is 6.5 K/km.
        z0 = 0
        T0 = 288.15
        gamma = 6.5e-3
        p0 = 101325.

        # Hydrostatic equation with linear temperature gradient.
        z = 1. / gamma * (T0 - T0 * numpy.exp(gamma * R / g * numpy.log(p / p0)))

    # Convert from m to flight level (ft).
    flightlevel = z * 0.0328083989502

    return flightlevel


def flightlevel2pressure_a(flightlevel):
    """
    Conversion of flight level (given in hft) to pressure (Pa) with
    hydrostatic equation, according to the profile of the ICAO
    standard atmosphere.

    Array version, the argument "flightlevel" must be a numpy array.

    Reference:
        For example, H. Kraus, Die Atmosphaere der Erde, Springer, 2001,
        470pp., Sections II.1.4. and II.6.1.2.

    Arguments:
        flightlevel -- numpy array of flight level in hft
    Returns:
        static pressure (Pa)
    """
    # Make sure flightlevel is a numpy array.
    if not isinstance(flightlevel, numpy.ndarray):
        raise ValueError("argument flightlevel must be a numpy array")

    # Convert flight level (ft) to m (1 ft = 30.48 cm; 1/0.3048m = 3.28...).
    z = flightlevel * 30.48

    if (z > 71000).any():
        raise ValueError("flight level to pressure conversion not "
                         "implemented for z > 71km")

    # g and R are used by all equations below.
    g = 9.80665
    R = 287.058

    # Initialize the return array.
    p = numpy.zeros(flightlevel.shape)

    # ICAO standard atmosphere between 0 and 11 km: T(z=0km) = 15 degC,
    # p(z=0km) = 1013.25 hPa. Temperature gradient is 6.5 K/km.
    indices = z <= 11000.
    z0 = 0
    T0 = 288.15
    gamma = 6.5e-3
    p0 = 101325.

    # Hydrostatic equation with linear temperature gradient.
    p[indices] = p0 * ((T0 - gamma * (z[indices] - z0)) / T0) ** (g / (gamma * R))

    # ICAO standard atmosphere between 11 and 20 km: T(z=11km) = -56.5 degC,
    # p(z=11km) = 226.32 hPa. Temperature is constant at -56.5 degC.
    indices = (z > 11000.) & (z <= 20000.)
    z0 = 11000.
    p0 = 22632.64
    T = 216.65

    # Hydrostatic equation with constant temperature profile.
    p[indices] = p0 * numpy.exp(-g * (z[indices] - z0) / (R * T))

    # ICAO standard atmosphere between 20 and 32 km: T(z=20km) = -56.5 degC,
    # p(z=20km) = 54.75 hPa. Temperature gradient is -1.0 K/km.
    indices = (z > 20000.) & (z <= 32000.)
    z0 = 20000.
    T0 = 216.65
    gamma = -1.0e-3
    p0 = 5475.16

    # Hydrostatic equation with linear temperature gradient.
    p[indices] = p0 * ((T0 - gamma * (z[indices] - z0)) / T0) ** (g / (gamma * R))

    # ICAO standard atmosphere between 32 and 47 km: T(z=32km) = -44.5 degC,
    # p(z=32km) = 8.68019 hPa. Temperature gradient is -2.8 K/km.
    indices = (z > 32000.) & (z <= 47000.)
    z0 = 32000.
    T0 = 228.66
    gamma = -2.8e-3
    p0 = 868.089

    # Hydrostatic equation with linear temperature gradient.
    p[indices] = p0 * ((T0 - gamma * (z[indices] - z0)) / T0) ** (g / (gamma * R))

    # ICAO standard atmosphere between 47 and 51 km: T(z=47km) = -2.5 degC,
    # p(z=47km) = 1.10906 hPa. Temperature is constant at -2.5 degC.
    indices = (z > 47000.) & (z <= 51000)
    z0 = 47000.
    T = 270.65
    p0 = 110.928

    # Hydrostatic equation with constant temperature profile.
    p[indices] = p0 * numpy.exp(-g * (z[indices] - z0) / (R * T))

    # ICAO standard atmosphere between 51 and 71 km: T(z=51km) = -2.5 degC,
    # p(z=71km) = 0.66939 hPa. Temperature gradient is 2.8 K/km.
    indices = (z > 51000.) & (z <= 71000)
    z0 = 51000.
    T0 = 270.65
    gamma = 2.8e-3
    p0 = 66.952

    # Hydrostatic equation with linear temperature gradient.
    p[indices] = p0 * ((T0 - gamma * (z[indices] - z0)) / T0) ** (g / (gamma * R))

    return p


def pressure2flightlevel_a(p):
    """
    Conversion of pressure (Pa) to flight level (hft) with
    hydrostatic equation, according to the profile of the ICAO
    standard atmosphere.

    Array version, the argument "p" must be a numpy array.

    Reference:
        For example, H. Kraus, Die Atmosphaere der Erde, Springer, 2001,
        470pp., Sections II.1.4. and II.6.1.2.

    Arguments:
        p -- numpy array of pressure (Pa)
        fake_above_32km -- compute values above 54.75 hPa (32km) with the
                           profile valid for 20..32km. WARNING: This gives
                           unphysical results. Use this option only for
                           testing purposes.
    Returns:
        flight level in hft
    """
    # Make sure p is a numpy array.
    if not isinstance(p, numpy.ndarray):
        raise ValueError("argument p must be a numpy array")

    # g and R are used by all equations below.
    g = 9.80665
    R = 287.058

    if (p < 3.9591).any():
        raise ValueError("pressure to flight level conversion not "
                         "implemented for z > 71km (p ~ 4 Pa)")

    # Initialize the return array.
    z = numpy.zeros_like(p)

    indices = p < 66.952
    # ICAO standard atmosphere between 51 and 71 km: T(z=51km) = -2.5 degC,
    # p(z=71km) = 0.66939 hPa. Temperature gradient is 2.8 K/km.
    z0 = 51000.
    T0 = 270.65
    gamma = 2.8e-3
    p0 = 66.952

    # Hydrostatic equation with linear temperature gradient.
    z[indices] = z0 + 1. / gamma * (T0 - T0 * numpy.exp(gamma * R / g * numpy.log(p[indices] / p0)))

    # ICAO standard atmosphere between 47 and 51km: T(z=47km) = -2.5 degC,
    # p(z=11km) = 1.10906 hPa. Temperature is constant at -2.5 degC.
    indices = (p >= 66.952) & (p < 110.928)
    z0 = 47000.
    p0 = 110.928
    T = 270.65

    # Hydrostatic equation with constant temperature profile.
    z[indices] = z0 - (R * T) / g * numpy.log(p[indices] / p0)

    # ICAO standard atmosphere between 32 and 47 km: T(z=32km) = -44.5 degC,
    # p(z=20km) = 8.68019 hPa. Temperature gradient is -2.8 K/km.
    indices = (p >= 110.928) & (p < 868.089)
    z0 = 32000.
    T0 = 228.66
    gamma = -2.8e-3
    p0 = 868.089

    # Hydrostatic equation with linear temperature gradient.
    z[indices] = z0 + 1. / gamma * (T0 - T0 * numpy.exp(gamma * R / g * numpy.log(p[indices] / p0)))

    # ICAO standard atmosphere between 20 and 32 km: T(z=20km) = -56.5 degC,
    # p(z=20km) = 54.75 hPa. Temperature gradient is -1.0 K/km.
    indices = (p >= 868.089) & (p < 5474.16)
    z0 = 20000.
    T0 = 216.65
    gamma = -1.0e-3
    p0 = 5475.16

    # Hydrostatic equation with linear temperature gradient.
    z[indices] = z0 + 1. / gamma * (T0 - T0 * numpy.exp(gamma * R / g * numpy.log(p[indices] / p0)))

    # ICAO standard atmosphere between 11 and 20 km: T(z=11km) = -56.5 degC,
    # p(z=11km) = 226.32 hPa. Temperature is constant at -56.5 degC.
    indices = (p >= 5474.16) & (p < 22632.)
    z0 = 11000.
    p0 = 22632.64
    T = 216.65

    # Hydrostatic equation with constant temperature profile.
    z[indices] = z0 - (R * T) / g * numpy.log(p[indices] / p0)

    # ICAO standard atmosphere between 0 and 11 km: T(z=0km) = 15 degC,
    # p(z=0km) = 1013.25 hPa. Temperature gradient is 6.5 K/km.
    indices = p >= 22632.
    z0 = 0
    T0 = 288.15
    gamma = 6.5e-3
    p0 = 101325.

    # Hydrostatic equation with linear temperature gradient.
    z[indices] = 1. / gamma * (T0 - T0 * numpy.exp(gamma * R / g * numpy.log(p[indices] / p0)))

    # Convert from m to flight level (ft).
    flightlevel = z * 0.0328083989502

    return flightlevel


def isa_temperature(flightlevel):
    """
    International standard atmosphere temperature at the given flight level.

    Reference:
        For example, H. Kraus, Die Atmosphaere der Erde, Springer, 2001,
        470pp., Sections II.1.4. and II.6.1.2.

    Arguments:
        flightlevel -- flight level in hft
    Returns:
        temperature (K)
    """
    # Convert flight level (ft) to m (1 ft = 30.48 cm; 1/0.3048m = 3.28...).
    z = flightlevel * 30.48

    if z <= 11000.:
        # ICAO standard atmosphere between 0 and 11 km: T(z=0km) = 15 degC,
        # p(z=0km) = 1013.25 hPa. Temperature gradient is 6.5 K/km.
        T0 = 288.15
        gamma = 6.5e-3
        return T0 - gamma * z

    elif z <= 20000.:
        # ICAO standard atmosphere between 11 and 20 km: T(z=11km) = -56.5 degC,
        # p(z=11km) = 226.32 hPa. Temperature is constant at -56.5 degC.
        T = 216.65
        return T

    elif z <= 32000.:
        # ICAO standard atmosphere between 20 and 32 km: T(z=20km) = -56.5 degC,
        # p(z=20km) = 54.75 hPa. Temperature gradient is -1.0 K/km.
        z0 = 20000.
        T0 = 216.65
        gamma = -1.0e-3
        return T0 - gamma * (z - z0)

    elif z <= 47000.:
        # ICAO standard atmosphere between 32 and 47 km: T(z=32km) = -44.5 degC,
        # p(z=32km) = 8.68019 hPa. Temperature gradient is -2.8 K/km.
        z0 = 32000.
        T0 = 228.65
        gamma = -2.8e-3
        return T0 - gamma * (z - z0)

    elif z <= 47820.070345213892438:
        # ICAO standard atmosphere between 47 and 47820.070345213892438 km: T(z=47km) = -2.5 degC,
        # p(z=47km) = 1.10906 hPa. Temperature is constant at -2.5 degC.
        T = 270.65
        return T

    else:
        raise ValueError("ISA temperature from flight level not "
                         "implemented for z > 71km")



# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.test_chat_manager
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    tests for chat_manager functionalities

    This file is part of mss.

    :copyright: Copyright 2020 Reimar Bauer
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import requests
import json
import sys
import time
import pytest

from werkzeug.urls import url_join
from PyQt5 import QtWidgets
from mslib.mscolab.models import User, MessageType, Message
from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.chat_manager import ChatManager
from mslib.msui.mscolab import MSSMscolabWindow
from mslib._tests.utils import mscolab_start_server


PORTS = list(range(9321, 9340))


@pytest.mark.skipif(os.name == "nt",
                    reason="multiprocessing needs currently start_method fork")
class Test_Chat_Manager(object):
    def setup(self):
        self.process, self.url, self.app, _, self.cm, self.fm = mscolab_start_server(PORTS)
        time.sleep(0.1)
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = MSSMscolabWindow(data_dir=mscolab_settings.MSCOLAB_DATA_DIR,
                                       mscolab_server_url=self.url)
        self.cm = ChatManager()
        self.room_name = "europe"
        data = {
            'email': 'a',
            'password': 'a'
        }
        r = requests.post(self.url + '/token', data=data)
        self.token = json.loads(r.text)['token']
        with self.app.app_context():
            self.user = User.query.filter_by(id=8).first()

        data = {
            "token": self.token,
            "path": self.room_name,
            "description": "test description"
        }
        url = url_join(self.url, 'create_project')
        requests.post(url, data=data)

    def teardown(self):
        if self.window.version_window:
            self.window.version_window.close()
        if self.window.conn:
            self.window.conn.disconnect()
        self.application.quit()
        QtWidgets.QApplication.processEvents()
        self.process.terminate()

    def test_add_message(self):
        with self.app.app_context():
            message = self.cm.add_message(self.user, 'some message', self.room_name, message_type=MessageType.TEXT,
                                          reply_id=None)
            assert message.text == 'some message'

    def test_edit_messages(self):
        with self.app.app_context():
            message = self.cm.add_message(self.user, 'some test message', self.room_name, message_type=MessageType.TEXT,
                                          reply_id=None)
            new_message_text = "Wonderland"
            self.cm.edit_message(message.id, new_message_text)
            message = Message.query.filter_by(id=message.id).first()
            assert message.text == new_message_text

    def test_delete_messages(self):
        with self.app.app_context():
            message = self.cm.add_message(self.user, 'some test example message',
                                          self.room_name, message_type=MessageType.TEXT,
                                          reply_id=None)
            assert 'some test example message' in message.text
            self.cm.delete_message(message.id)
            message = Message.query.filter(Message.id == message.id).first()
            assert message is None



# -*- coding: utf-8 -*-
"""

    mslib.msui.editor
    ~~~~~~~~~~~~~~~~~~~~~~

    config editor for mss_settings.json.

    This file is part of mss.

    :copyright: Copyright 2020 Vaibhav Mehra <veb7vmehra@gmail.com>
    :copyright: Copyright 2020-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import fs
import logging
import json

from mslib.msui.mss_qt import get_open_filename, get_save_filename
from PyQt5 import QtWidgets, QtGui, QtCore, QtPrintSupport
from mslib.msui import constants
from mslib.msui.constants import MSS_CONFIG_PATH
from mslib.msui.icons import icons


class EditorMainWindow(QtWidgets.QMainWindow):
    name = "MSSEditor"
    identifier = None

    viewCloses = QtCore.pyqtSignal(name="viewCloses")
    restartApplication = QtCore.pyqtSignal(name="restartApplication")

    def __init__(self, parent=None):
        super(EditorMainWindow, self).__init__(parent)
        self.path = None

        self.file_content = None
        self.layout = QtWidgets.QVBoxLayout()
        # Could also use a QTextEdit and set self.editor.setAcceptRichText(False)
        self.editor = QtWidgets.QPlainTextEdit()

        # Load mss_settings.json (if already exists)
        self.path = constants.CACHED_CONFIG_FILE
        if self.path:
            dir_name, file_name = fs.path.split(self.path)
            with fs.open_fs(dir_name) as _fs:
                if _fs.exists(file_name):
                    self.file_content = _fs.readtext(file_name)
                    self.editor.setPlainText(self.file_content)
                    self.update_title()
        self.last_saved = self.editor.toPlainText()

        # Setup the QTextEdit editor configuration
        fixedfont = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
        fixedfont.setPointSize(12)
        self.editor.setFont(fixedfont)

        self.layout.addWidget(self.editor)

        self.container = QtWidgets.QWidget()
        self.container.setLayout(self.layout)
        self.setCentralWidget(self.container)

        self.status = QtWidgets.QStatusBar()
        self.setStatusBar(self.status)

        self.file_toolbar = QtWidgets.QToolBar("File")
        self.file_toolbar.setIconSize(QtCore.QSize(14, 14))
        self.addToolBar(self.file_toolbar)
        self.file_menu = self.menuBar().addMenu("&File")

        self.open_file_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                                    'Folder-new.svg')), "Open file...", self)
        self.open_file_action.setStatusTip("Open file")
        self.open_file_action.triggered.connect(self.file_open)
        self.file_menu.addAction(self.open_file_action)
        self.file_toolbar.addAction(self.open_file_action)

        self.save_file_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                                    'Document-save.svg')), "Save and Quit", self)
        self.save_file_action.setStatusTip("Save current page")
        self.save_file_action.triggered.connect(self.file_save_and_quit)
        self.file_menu.addAction(self.save_file_action)
        self.file_toolbar.addAction(self.save_file_action)

        self.saveas_file_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                                      'Document-save-as.svg')), "Save As...", self)
        self.saveas_file_action.setStatusTip("Save current page to specified file")
        self.saveas_file_action.triggered.connect(self.file_saveas)
        self.file_menu.addAction(self.saveas_file_action)
        self.file_toolbar.addAction(self.saveas_file_action)

        self.print_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                                'Document-print.svg')), "Print...", self)
        self.print_action.setStatusTip("Print current page")
        self.print_action.triggered.connect(self.file_print)
        self.file_menu.addAction(self.print_action)
        self.file_toolbar.addAction(self.print_action)

        self.quit_action = QtWidgets.QAction("Quit", self)
        self.quit_action.triggered.connect(self.close)
        self.file_menu.addAction(self.quit_action)

        self.edit_toolbar = QtWidgets.QToolBar("Edit")
        self.edit_toolbar.setIconSize(QtCore.QSize(16, 16))
        self.addToolBar(self.edit_toolbar)
        self.edit_menu = self.menuBar().addMenu("&Edit")

        self.undo_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                               'Edit-undo.svg')), "Undo", self)
        self.undo_action.setStatusTip("Undo last change")
        self.undo_action.triggered.connect(self.editor.undo)
        self.edit_menu.addAction(self.undo_action)

        self.redo_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                               'Edit-redo.svg')), "Redo", self)
        self.redo_action.setStatusTip("Redo last change")
        self.redo_action.triggered.connect(self.editor.redo)
        self.edit_toolbar.addAction(self.redo_action)
        self.edit_menu.addAction(self.redo_action)

        self.edit_menu.addSeparator()

        self.cut_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                              'Edit-cut.svg')), "Cut", self)
        self.cut_action.setStatusTip("Cut selected text")
        self.cut_action.triggered.connect(self.editor.cut)
        self.edit_toolbar.addAction(self.cut_action)
        self.edit_menu.addAction(self.cut_action)

        self.copy_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                               'Edit-copy.svg')), "Copy", self)
        self.copy_action.setStatusTip("Copy selected text")
        self.copy_action.triggered.connect(self.editor.copy)
        self.edit_toolbar.addAction(self.copy_action)
        self.edit_menu.addAction(self.copy_action)

        self.paste_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                                'Edit-paste.svg')), "Paste", self)
        self.paste_action.setStatusTip("Paste from clipboard")
        self.paste_action.triggered.connect(self.editor.paste)
        self.edit_toolbar.addAction(self.paste_action)
        self.edit_menu.addAction(self.paste_action)

        self.select_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                                 'Edit-select-all.svg')), "Select all", self)
        self.select_action.setStatusTip("Select all text")
        self.select_action.triggered.connect(self.editor.selectAll)
        self.edit_menu.addAction(self.select_action)

        self.edit_menu.addSeparator()

        self.wrap_action = QtWidgets.QAction(QtGui.QIcon(icons('config_editor',
                                                               'Go-next.svg')), "Wrap text to window", self)
        self.wrap_action.setStatusTip("Toggle wrap text to window")
        self.wrap_action.setCheckable(True)
        self.wrap_action.setChecked(True)
        self.wrap_action.triggered.connect(self.edit_toggle_wrap)
        self.edit_menu.addAction(self.wrap_action)
        self.update_title()
        self.show()

    def file_open(self):
        file_path = get_open_filename(self, "Open file", MSS_CONFIG_PATH, "Text documents (*.json)")
        if file_path is not None:
            file_name = fs.path.basename(file_path)
            with fs.open_fs(fs.path.dirname(file_path)) as file_dir:
                self.file_content = file_dir.readtext(file_name)
                self.last_saved = self.file_content
                self.editor.setPlainText(self.file_content)
                self.update_title()

    def check_modified(self):
        if self.last_saved != self.editor.toPlainText():
            return True
        return False

    def check_json(self):
        try:
            json.loads(self.editor.toPlainText())
        except json.JSONDecodeError as ex:
            QtWidgets.QMessageBox.warning(
                self, self.tr("Mission Support System"),
                self.tr("This JSON file contains Syntax errors and will not be saved.\n") + str(ex))
            return False
        return True

    def file_save_and_quit(self):
        if self.check_modified():
            if self.check_json():
                self._save_to_path(self.path)

                ret = QtWidgets.QMessageBox.warning(
                    self, self.tr("Mission Support System"),
                    self.tr("Do you want to restart the application?\n"
                            "(This is necessary to apply changes)"),
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                    QtWidgets.QMessageBox.No)
                if ret == QtWidgets.QMessageBox.Yes:
                    self.restartApplication.emit()
        self.close()

    def file_saveas(self):
        if self.check_json():
            default_filename = constants.CACHED_CONFIG_FILE
            path = get_save_filename(
                self, "Save file", default_filename, "Text documents (*.json)")
            # If dialog is cancelled, will return ''
            if path:
                self._save_to_path(path)

    def _save_to_path(self, filename):
        logging.debug("save config file to: %s", filename)
        text = self.editor.toPlainText()
        self.last_saved = text
        dir_name, file_name = fs.path.split(filename)
        with fs.open_fs(dir_name) as _fs:
            _fs.writetext(file_name, text)
        self.update_title()

    def file_print(self):
        dlg = QtPrintSupport.QPrintDialog()
        if dlg.exec_():
            self.editor.print_(dlg.printer())

    def update_title(self):
        self.setWindowTitle(f"{fs.path.basename(self.path)} - Config-Settings" if self.path else "Untitled")

    def edit_toggle_wrap(self):
        self.editor.setLineWrapMode(1 if self.editor.lineWrapMode() == 0 else 0)

    def closeEvent(self, event):
        if self.check_modified() and self.check_json():
            ret = QtWidgets.QMessageBox.question(
                self, self.tr("Mission Support System"),
                self.tr("Save Changes to default mss_settings.json?\n"
                        "You need to restart the gui for changes to take effect."),
                QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)

            if ret == QtWidgets.QMessageBox.Yes:
                self.file_save_and_quit()

        self.viewCloses.emit()
        event.accept()



# -*- coding: utf-8 -*-
"""

    mslib.msui.mss_qt
    ~~~~~~~~~~~~~~~~~

    This module helps with qt

    This file is part of mss.

    :copyright: Copyright 2017-2018 Joern Ungermann, Reimar Bauer
    :copyright: Copyright 2017-2021 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import importlib
import logging
import os
import platform
import sys
import traceback

from fslib.fs_filepicker import getSaveFileName, getOpenFileName, getExistingDirectory
from PyQt5 import QtCore, QtWidgets  # noqa

from mslib.utils import config_loader, FatalUserError


def get_open_filename_qt(*args):
    filename = QtWidgets.QFileDialog.getOpenFileName(*args)
    return filename[0] if isinstance(filename, tuple) else str(filename)


def get_open_filenames_qt(*args):
    """
    To select multiple files simultaneously
    """
    filenames = QtWidgets.QFileDialog.getOpenFileNames(*args)
    return filenames[0] if isinstance(filenames, tuple) else str(filenames)


def get_save_filename_qt(*args):
    filename = QtWidgets.QFileDialog.getSaveFileName(*args)
    return filename[0] if isinstance(filename, tuple) else str(filename)


def get_existing_directory_qt(*args):
    dirname = QtWidgets.QFileDialog.getExistingDirectory(*args)
    return dirname[0] if isinstance(dirname, tuple) else str(dirname)


def get_pickertype(tag, typ):
    default = config_loader(dataset="filepicker_default")
    if typ is None:
        if tag is None:
            typ = default
        else:
            typ = config_loader(dataset=tag)
    return typ


def get_open_filename(parent, title, dirname, filt, pickertag=None, pickertype=None):
    pickertype = get_pickertype(pickertag, pickertype)
    if pickertype == "fs":
        # fs filepicker takes file filters as a list
        if not isinstance(filt, list):
            filt = filt.split(';;')
        filename = getOpenFileName(parent, dirname, filt, title="Import Flight Track")
    elif pickertype in ["qt", "default"]:
        # qt filepicker takes file filters separated by ';;'
        filename = get_open_filename_qt(parent, title, os.path.expanduser(dirname), filt)
    else:
        raise FatalUserError(f"Unknown file picker type '{pickertype}'.")
    logging.debug("Selected '%s'", filename)
    if filename == "":
        filename = None
    return filename


def get_open_filenames(parent, title, dirname, filt, pickertag=None, pickertype=None):
    """
    Opens multiple files simultaneously
    Currently implemented only in kmloverlay_dockwidget.py
    """
    pickertype = get_pickertype(pickertag, pickertype)
    if pickertype in ["qt", "default"]:
        filename = get_open_filenames_qt(parent, title, os.path.expanduser(dirname), filt)
    else:
        raise FatalUserError(f"Unknown file picker type '{pickertype}'.")
    logging.debug("Selected '%s'", filename)
    if filename == "":
        filename = None
    return filename


def get_save_filename(parent, title, filename, filt, pickertag=None, pickertype=None):
    pickertype = get_pickertype(pickertag, pickertype)
    if pickertype == "fs":
        dirname, filename = os.path.split(filename)
        filename = getSaveFileName(
            parent, dirname, filt, title=title, default_filename=filename, show_save_action=True)
    elif pickertype in ["qt", "default"]:
        filename = get_save_filename_qt(parent, title, os.path.expanduser(filename), filt)
    else:
        raise FatalUserError(f"Unknown file picker type '{pickertype}'.")
    logging.debug("Selected '%s'", filename)
    if filename == "":
        filename = None
    return filename


def get_existing_directory(parent, title, defaultdir, pickertag=None, pickertype=None):
    pickertype = get_pickertype(pickertag, pickertype)
    if pickertype == "fs":
        dirname = getExistingDirectory(parent, title=title, fs_url=defaultdir)[0]
    elif pickertype in ["qt", "default"]:
        dirname = get_existing_directory_qt(parent, title, defaultdir)
    else:
        raise FatalUserError(f"Unknown file picker type '{pickertype}'.")
    logging.debug("Selected '%s'", dirname)
    if dirname == "":
        dirname = None
    return dirname


def variant_to_string(variant):
    if isinstance(variant, QtCore.QVariant):
        return str(variant.value())
    return str(variant)


def variant_to_float(variant, locale=QtCore.QLocale()):
    if isinstance(variant, QtCore.QVariant):
        value = variant.value()
    else:
        value = variant

    if isinstance(value, (int, float)):
        return value
    try:
        float_value, ok = locale.toDouble(value)
        if not ok:
            raise ValueError
    except TypeError:  # neither float nor string, try Python conversion
        logging.error("Unexpected type in float conversion: %s=%s",
                      type(value), value)
        float_value = float(value)
    return float_value


# Import all Dialogues from the proper module directory.
for mod in [
        "ui_about_dialog",
        "ui_hexagon_dockwidget",
        "ui_kmloverlay_dockwidget",
        "ui_customize_kml",
        "ui_mainwindow",
        "ui_performance_dockwidget",
        "ui_remotesensing_dockwidget",
        "ui_satellite_dockwidget",
        "ui_sideview_options",
        "ui_sideview_window",
        "ui_tableview_window",
        "ui_topview_mapappearance",
        "ui_topview_window",
        "ui_wms_capabilities",
        "ui_wms_dockwidget",
        "ui_wms_password_dialog",
        "ui_wms_multilayers"]:
    globals()[mod] = importlib.import_module("mslib.msui.qt5." + mod)

# to store config by QSettings
QtCore.QCoreApplication.setOrganizationName("mss")


# PyQt5 silently aborts on a Python Exception
def excepthook(type_, value, traceback_):
    """
    This dumps the error to console, logging (i.e. logfile), and tries to open a MessageBox for GUI users.
    """
    import mslib
    import mslib.utils
    tb = "".join(traceback.format_exception(type_, value, traceback_))
    traceback.print_exception(type_, value, traceback_)
    logging.critical("MSS Version: %s", mslib.__version__)
    logging.critical("Python Version: %s", sys.version)
    logging.critical("Platform: %s (%s)", platform.platform(), platform.architecture())
    logging.critical("Fatal error: %s", tb)

    if type_ is mslib.utils.FatalUserError:
        QtWidgets.QMessageBox.critical(
            None, "fatal error",
            f"Fatal user error in MSS {mslib.__version__} on {platform.platform()}\n"
            f"Python {sys.version}\n"
            f"\n"
            f"{value}")
    else:
        QtWidgets.QMessageBox.critical(
            None, "fatal error",
            f"Fatal error in MSS {mslib.__version__} on {platform.platform()}\n"
            f"Python {sys.version}\n"
            f"\n"
            f"Please report bugs in MSS to https://github.com/Open-MSS/MSS\n"
            f"\n"
            f"Information about the fatal error:\n"
            f"\n"
            f"{tb}")


sys.excepthook = excepthook
